import logging
from typing import Any, Callable, Optional

import numpy as np
from numpy.typing import ArrayLike

logger = logging.getLogger(__name__)


def threshold_predictions(y: np.ndarray,
                          threshold: Optional[float] = None) -> np.ndarray:
    """Threshold predictions from classification model.

    Parameters
    ----------
    y: np.ndarray
        Must have shape `(N, n_classes)` and be class probabilities.
    threshold: float, default None
        The threshold probability for the positive class. Note that this
        threshold will only be applied for binary classifiers (where
        `n_classes==2`). If specified for multiclass problems, or if
        `threshold` is None, the threshold is ignored and argmax(y) is
        returned.

    Returns
    -------
    y_out: np.ndarray
        A numpy array of shape `(N,)` with class predictions as integers ranging from 0
        to `n_classes-1`.
    """
    if not isinstance(y, np.ndarray) or not len(y.shape) == 2:
        raise ValueError("y must be a ndarray of shape (N, n_classes)")
    N = y.shape[0]
    n_classes = y.shape[1]
    if n_classes != 2 or threshold is None:
        return np.argmax(y, axis=1)
    else:
        return np.where(y[:, 1] >= threshold, np.ones(N), np.zeros(N))


def normalize_weight_shape(w: Optional[np.ndarray], n_samples: int,
                           n_tasks: int) -> np.ndarray:
    """A utility function to correct the shape of the weight array.

    This utility function is used to normalize the shapes of a given
    weight array.

    Parameters
    ----------
    w: np.ndarray
        `w` can be `None` or a scalar or a `np.ndarray` of shape
        `(n_samples,)` or of shape `(n_samples, n_tasks)`. If `w` is a
        scalar, it's assumed to be the same weight for all samples/tasks.
    n_samples: int
        The number of samples in the dataset. If `w` is not None, we should
        have `n_samples = w.shape[0]` if `w` is a ndarray
    n_tasks: int
        The number of tasks. If `w` is 2d ndarray, then we should have
        `w.shape[1] == n_tasks`.

    Examples
    --------
    >>> import numpy as np
    >>> w_out = normalize_weight_shape(None, n_samples=10, n_tasks=1)
    >>> (w_out == np.ones((10, 1))).all()
    True

    Returns
    -------
    w_out: np.ndarray
        Array of shape `(n_samples, n_tasks)`
    """
    if w is None:
        w_out = np.ones((n_samples, n_tasks))
    elif isinstance(w, np.ndarray):
        if len(w.shape) == 0:
            # scalar case
            w_out = w * np.ones((n_samples, n_tasks))
        elif len(w.shape) == 1:
            if len(w) != n_samples:
                raise ValueError("Length of w isn't n_samples")
            # per-example case
            # This is a little arcane but it repeats w across tasks.
            w_out = np.tile(w, (n_tasks, 1)).T
        elif len(w.shape) == 2:
            if w.shape == (n_samples, 1):
                # If w.shape == (n_samples, 1) handle it as 1D
                w = np.squeeze(w, axis=1)
                w_out = np.tile(w, (n_tasks, 1)).T
            elif w.shape != (n_samples, n_tasks):
                raise ValueError(
                    "Shape for w doens't match (n_samples, n_tasks)")
            else:
                # w.shape == (n_samples, n_tasks)
                w_out = w
        else:
            raise ValueError("w must be of dimension 1, 2, or 3")
    else:
        # scalar case
        w_out = w * np.ones((n_samples, n_tasks))
    return w_out


def normalize_labels_shape(y: np.ndarray,
                           mode: Optional[str] = None,
                           n_tasks: Optional[int] = None,
                           n_classes: Optional[int] = None) -> np.ndarray:
    """A utility function to correct the shape of the labels.

    Parameters
    ----------
    y: np.ndarray
        `y` is an array of shape `(N,)` or `(N, n_tasks)` or `(N, n_tasks, 1)`.
    mode: str, default None
        If `mode` is "classification" or "regression", attempts to apply
        data transformations.
    n_tasks: int, default None
        The number of tasks this class is expected to handle.
    n_classes: int, default None
        If specified use this as the number of classes. Else will try to
        impute it as `n_classes = max(y) + 1` for arrays and as
        `n_classes=2` for the case of scalars. Note this parameter only
        has value if `mode=="classification"`

    Returns
    -------
    y_out: np.ndarray
        If `mode=="classification"`, `y_out` is an array of shape `(N,
        n_tasks, n_classes)`. If `mode=="regression"`, `y_out` is an array
        of shape `(N, n_tasks)`.
    """
    if n_tasks is None:
        raise ValueError("n_tasks must be specified")
    if mode not in ["classification", "regression"]:
        raise ValueError("mode must be either classification or regression.")
    if mode == "classification" and n_classes is None:
        raise ValueError("n_classes must be specified")
    if not isinstance(y, np.ndarray):
        raise ValueError("y must be a np.ndarray")
    # Handle n_classes/n_task shape ambiguity
    if mode == "classification" and len(y.shape) == 2:
        if n_classes == y.shape[1] and n_tasks != 1 and n_classes != n_tasks:
            raise ValueError("Shape of input doesn't match expected n_tasks=1")
        elif n_classes == y.shape[1] and n_tasks == 1:
            # Add in task dimension
            y = np.expand_dims(y, 1)
    if len(y.shape) == 1 and n_tasks != 1:
        raise ValueError("n_tasks must equal 1 for a 1D set of labels.")
    if (len(y.shape) == 2 or len(y.shape) == 3) and n_tasks != y.shape[1]:
        raise ValueError("Shape of input doesn't match expected n_tasks=%d" %
                         n_tasks)
    if len(y.shape) >= 4:
        raise ValueError(
            "Labels y must be a float scalar or a ndarray of shape `(N,)` or "
            "`(N, n_tasks)` or `(N, n_tasks, 1)` for regression problems and "
            "of shape `(N,)` or `(N, n_tasks)` or `(N, n_tasks, 1)` for classification problems"
        )
    if len(y.shape) == 1:
        # Insert a task dimension (we know n_tasks=1 from above0
        y_out = np.expand_dims(y, 1)
    elif len(y.shape) == 2:
        y_out = y
    elif len(y.shape) == 3:
        # If 3D and last dimension isn't 1, assume this is one-hot encoded and return as-is.
        if y.shape[-1] != 1:
            return y
        y_out = np.squeeze(y, axis=-1)
    # Handle classification. We need to convert labels into one-hot representation.
    if mode == "classification":
        all_y_task = []
        for task in range(n_tasks):
            y_task = y_out[:, task]
            # check whether n_classes is int or not
            assert isinstance(n_classes, int)
            y_hot = to_one_hot(y_task, n_classes=n_classes)
            y_hot = np.expand_dims(y_hot, 1)
            all_y_task.append(y_hot)
        y_out = np.concatenate(all_y_task, axis=1)
    return y_out


def normalize_prediction_shape(y: np.ndarray,
                               mode: Optional[str] = None,
                               n_tasks: Optional[int] = None,
                               n_classes: Optional[int] = None):
    """A utility function to correct the shape of provided predictions.

    The metric computation classes expect that inputs for classification
    have the uniform shape `(N, n_tasks, n_classes)` and inputs for
    regression have the uniform shape `(N, n_tasks)`. This function
    normalizes the provided input array to have the desired shape.

    Examples
    --------
    >>> import numpy as np
    >>> y = np.random.rand(10)
    >>> y_out = normalize_prediction_shape(y, "regression", n_tasks=1)
    >>> y_out.shape
    (10, 1)

    Parameters
    ----------
    y: np.ndarray
        If `mode=="classification"`, `y` is an array of shape `(N,)` or
        `(N, n_tasks)` or `(N, n_tasks, n_classes)`. If
        `mode=="regression"`, `y` is an array of shape `(N,)` or `(N,
        n_tasks)`or `(N, n_tasks, 1)`.
    mode: str, default None
        If `mode` is "classification" or "regression", attempts to apply
        data transformations.
    n_tasks: int, default None
        The number of tasks this class is expected to handle.
    n_classes: int, default None
        If specified use this as the number of classes. Else will try to
        impute it as `n_classes = max(y) + 1` for arrays and as
        `n_classes=2` for the case of scalars. Note this parameter only
        has value if `mode=="classification"`

    Returns
    -------
    y_out: np.ndarray
      If `mode=="classification"`, `y_out` is an array of shape `(N,
      n_tasks, n_classes)`. If `mode=="regression"`, `y_out` is an array
      of shape `(N, n_tasks)`.
    """
    if n_tasks is None:
        raise ValueError("n_tasks must be specified")
    if mode == "classification" and n_classes is None:
        raise ValueError("n_classes must be specified")
    if not isinstance(y, np.ndarray):
        raise ValueError("y must be a np.ndarray")
    # Handle n_classes/n_task shape ambiguity
    if mode == "classification" and len(y.shape) == 2:
        if n_classes == y.shape[1] and n_tasks != 1 and n_classes != n_tasks:
            raise ValueError("Shape of input doesn't match expected n_tasks=1")
        elif n_classes == y.shape[1] and n_tasks == 1:
            # Add in task dimension
            y = np.expand_dims(y, 1)
    if (len(y.shape) == 2 or len(y.shape) == 3) and n_tasks != y.shape[1]:
        raise ValueError("Shape of input doesn't match expected n_tasks=%d" %
                         n_tasks)
    if len(y.shape) >= 4:
        raise ValueError(
            "Predictions y must be a float scalar or a ndarray of shape `(N,)` or "
            "`(N, n_tasks)` or `(N, n_tasks, 1)` for regression problems and "
            "of shape `(N,)` or `(N, n_tasks)` or `(N, n_tasks, n_classes)` for classification problems"
        )
    if mode == "classification":
        if n_classes is None:
            raise ValueError("n_classes must be specified.")
        if len(y.shape) == 1 or len(y.shape) == 2:
            # Make everything 2D so easy to handle
            if len(y.shape) == 1:
                y = y[:, np.newaxis]
            # Handle each task separately.
            all_y_task = []
            for task in range(n_tasks):
                y_task = y[:, task]
                if len(np.unique(y_task)) > n_classes:
                    # Handle continuous class probabilites of positive class for binary
                    if n_classes > 2:
                        raise ValueError(
                            "Cannot handle continuous probabilities for multiclass problems."
                            "Need a per-class probability")
                    # Fill in class 0 probabilities
                    y_task = np.array([1 - y_task, y_task]).T
                    # Add a task dimension to concatenate on
                    y_task = np.expand_dims(y_task, 1)
                    all_y_task.append(y_task)
                else:
                    # Handle binary labels
                    # make y_hot of shape (N, n_classes)
                    y_task = to_one_hot(y_task, n_classes=n_classes)
                    # Add a task dimension to concatenate on
                    y_task = np.expand_dims(y_task, 1)
                    all_y_task.append(y_task)
            y_out = np.concatenate(all_y_task, axis=1)
        elif len(y.shape) == 3:
            y_out = y
    elif mode == "regression":
        if len(y.shape) == 1:
            # Insert a task dimension
            y_out = np.expand_dims(y, 1)
        elif len(y.shape) == 2:
            y_out = y
        elif len(y.shape) == 3:
            if y.shape[-1] != 1:
                raise ValueError(
                    "y must be a float scalar or a ndarray of shape `(N,)` or "
                    "`(N, n_tasks)` or `(N, n_tasks, 1)` for regression problems."
                )
            y_out = np.squeeze(y, axis=-1)
    else:
        raise ValueError("mode must be either classification or regression.")
    return y_out


def handle_classification_mode(
        y: np.ndarray,
        classification_handling_mode: Optional[str],
        threshold_value: Optional[float] = None) -> np.ndarray:
    """Handle classification mode.

    Transform predictions so that they have the correct classification mode.

    Parameters
    ----------
    y: np.ndarray
        Must be of shape `(N, n_tasks, n_classes)`
    classification_handling_mode: str, default None
        DeepChem models by default predict class probabilities for
        classification problems. This means that for a given singletask
        prediction, after shape normalization, the DeepChem prediction will be a
        numpy array of shape `(N, n_classes)` with class probabilities.
        `classification_handling_mode` is a string that instructs this method
        how to handle transforming these probabilities. It can take on the
        following values:
        - None: default value. Pass in `y_pred` directy into `self.metric`.
        - "threshold": Use `threshold_predictions` to threshold `y_pred`. Use
          `threshold_value` as the desired threshold.
        - "threshold-one-hot": Use `threshold_predictions` to threshold `y_pred`
          using `threshold_values`, then apply `to_one_hot` to output.
    threshold_value: float, default None
        If set, and `classification_handling_mode` is "threshold" or
        "threshold-one-hot" apply a thresholding operation to values with this
        threshold. This option isj only sensible on binary classification tasks.
        If float, this will be applied as a binary classification value.

    Returns
    -------
    y_out: np.ndarray
        If `classification_handling_mode` is "direct", then of shape `(N, n_tasks, n_classes)`.
        If `classification_handling_mode` is "threshold", then of shape `(N, n_tasks)`.
        If `classification_handling_mode is "threshold-one-hot", then of shape `(N, n_tasks, n_classes)"
    """
    if len(y.shape) != 3:
        raise ValueError("y must be of shape (N, n_tasks, n_classes)")
    N, n_tasks, n_classes = y.shape
    if classification_handling_mode == "direct":
        return y
    elif classification_handling_mode == "threshold":
        thresholded = []
        for task in range(n_tasks):
            task_array = y[:, task, :]
            # Now of shape (N,)
            task_array = threshold_predictions(task_array, threshold_value)
            # Now of shape (N, 1)
            task_array = np.expand_dims(task_array, 1)
            thresholded.append(task_array)
        # Returns shape (N, n_tasks)
        return np.concatenate(thresholded, axis=1)
    elif classification_handling_mode == "threshold-one-hot":
        thresholded = []
        for task in range(n_tasks):
            task_array = y[:, task, :]
            # Now of shape (N,)
            task_array = threshold_predictions(task_array, threshold_value)
            # Now of shape (N, n_classes)
            task_array = to_one_hot(task_array, n_classes=n_classes)
            # Now of shape (N, 1, n_classes)
            task_array = np.expand_dims(task_array, 1)
            thresholded.append(task_array)
        # Returns shape (N, n_tasks, n_classes)
        return np.concatenate(thresholded, axis=1)
    else:
        raise ValueError(
            "classification_handling_mode must be one of direct, threshold, threshold-one-hot"
        )


def to_one_hot(y: np.ndarray, n_classes: int = 2) -> np.ndarray:
    """Transforms label vector into one-hot encoding.

    Turns y into vector of shape `(N, n_classes)` with a one-hot
    encoding. Assumes that `y` takes values from `0` to `n_classes - 1`.

    Parameters
    ----------
    y: np.ndarray
        A vector of shape `(N,)` or `(N, 1)`
    n_classes: int, default 2
        If specified use this as the number of classes. Else will try to
        impute it as `n_classes = max(y) + 1` for arrays and as
        `n_classes=2` for the case of scalars. Note this parameter only
        has value if `mode=="classification"`

    Returns
    -------
    np.ndarray
        A numpy array of shape `(N, n_classes)`.
    """
    if len(y.shape) > 2:
        raise ValueError("y must be a vector of shape (N,) or (N, 1)")
    if len(y.shape) == 2 and y.shape[1] != 1:
        raise ValueError("y must be a vector of shape (N,) or (N, 1)")
    if len(np.unique(y)) > n_classes:
        raise ValueError("y has more than n_class unique elements.")
    N = np.shape(y)[0]
    y_hot = np.zeros((N, n_classes))
    y_hot[np.arange(N), y.astype(np.int64)] = 1
    return y_hot


def from_one_hot(y: np.ndarray, axis: int = 1) -> np.ndarray:
    """Transforms label vector from one-hot encoding.

    Parameters
    ----------
    y: np.ndarray
        A vector of shape `(n_samples, num_classes)`
    axis: int, optional (default 1)
        The axis with one-hot encodings to reduce on.

    Returns
    -------
    np.ndarray
        A numpy array of shape `(n_samples,)`
    """
    return np.argmax(y, axis=axis)


class Metric(object):
    """Wrapper class for computing user-defined metrics.

    The `Metric` class provides a wrapper for standardizing the API
    around different classes of metrics that may be useful for DeepChem
    models. The implementation provides a few non-standard conveniences
    such as built-in support for multitask and multiclass metrics.

    There are a variety of different metrics this class aims to support.
    Metrics for classification and regression that assume that values to
    compare are scalars are supported.

    At present, this class doesn't support metric computation on models
    which don't present scalar outputs. For example, if you have a
    generative model which predicts images or molecules, you will need
    to write a custom evaluation and metric setup.
    """

    def __init__(self,
                 metric: Callable[..., float],
                 task_averager: Optional[Callable[..., Any]] = None,
                 name: Optional[str] = None,
                 threshold: Optional[float] = None,
                 mode: Optional[str] = None,
                 n_tasks: Optional[int] = None,
                 classification_handling_mode: Optional[str] = None,
                 threshold_value: Optional[float] = None):
        """
        Parameters
        ----------
        metric: function
            Function that takes args y_true, y_pred (in that order) and
            computes desired score. If sample weights are to be considered,
            `metric` may take in an additional keyword argument
            `sample_weight`.
        task_averager: function, default None
            If not None, should be a function that averages metrics across
            tasks.
        name: str, default None
            Name of this metric
        threshold: float, default None (DEPRECATED)
            Used for binary metrics and is the threshold for the positive
            class.
        mode: str, default None
            Should usually be "classification" or "regression."
        n_tasks: int, default None
            The number of tasks this class is expected to handle.
        classification_handling_mode: str, default None
            DeepChem models by default predict class probabilities for
            classification problems. This means that for a given singletask
            prediction, after shape normalization, the DeepChem labels and prediction will be
            numpy arrays of shape `(n_samples, n_tasks, n_classes)` with class probabilities.
            `classification_handling_mode` is a string that instructs this method
            how to handle transforming these probabilities. It can take on the
            following values:
            - "direct": Pass `y_true` and `y_pred` directy into `self.metric`.
            - "threshold": Use `threshold_predictions` to threshold `y_true` and `y_pred`.
              Use `threshold_value` as the desired threshold. This converts them into
              arrays of shape `(n_samples, n_tasks)`, where each element is a class index.
            - "threshold-one-hot": Use `threshold_predictions` to threshold `y_true` and `y_pred`
              using `threshold_values`, then apply `to_one_hot` to output.
            - None: Select a mode automatically based on the metric.
        threshold_value: float, default None
            If set, and `classification_handling_mode` is "threshold" or
            "threshold-one-hot", apply a thresholding operation to values with this
            threshold. This option is only sensible on binary classification tasks.
            For multiclass problems, or if `threshold_value` is None, argmax() is used
            to select the highest probability class for each task.
        """
        if threshold is not None:
            logger.warn(
                "threshold is deprecated and will be removed in a future version of DeepChem."
                "Set threshold in compute_metric instead.")

        self.metric = metric
        if task_averager is None:
            self.task_averager = np.mean
        else:
            self.task_averager = task_averager
        if name is None:
            if task_averager is None:
                if hasattr(self.metric, '__name__'):
                    self.name = self.metric.__name__
                else:
                    self.name = "unknown metric"
            else:
                if hasattr(self.metric, '__name__'):
                    self.name = task_averager.__name__ + "-" + self.metric.__name__
                else:
                    self.name = "unknown metric"
        else:
            self.name = name

        if mode is None:
            # These are some smart defaults
            if self.metric.__name__ in [
                    "roc_auc_score", "matthews_corrcoef", "recall_score",
                    "accuracy_score", "kappa_score", "cohen_kappa_score",
                    "precision_score", "precision_recall_curve",
                    "balanced_accuracy_score", "prc_auc_score", "f1_score",
                    "bedroc_score", "jaccard_score", "jaccard_index",
                    "pixel_error"
            ]:
                mode = "classification"
            elif self.metric.__name__ in [
                    "pearson_r2_score", "r2_score", "mean_squared_error",
                    "mean_absolute_error", "rms_score", "mae_score", "pearsonr",
                    "concordance_index"
            ]:
                mode = "regression"
            else:
                raise ValueError(
                    "Please specify the mode of this metric. mode must be 'regression' or 'classification'"
                )
        if mode == "classification":
            if classification_handling_mode is None:
                # These are some smart defaults corresponding to sklearn's required
                # behavior
                if self.metric.__name__ in [
                        "matthews_corrcoef", "cohen_kappa_score", "kappa_score",
                        "balanced_accuracy_score", "recall_score",
                        "jaccard_score", "jaccard_index", "pixel_error",
                        "f1_score"
                ]:
                    classification_handling_mode = "threshold"
                elif self.metric.__name__ in [
                        "accuracy_score", "precision_score", "bedroc_score"
                ]:
                    classification_handling_mode = "threshold-one-hot"
                elif self.metric.__name__ in [
                        "roc_auc_score", "prc_auc_score",
                        "precision_recall_curve"
                ]:
                    classification_handling_mode = "direct"
            if classification_handling_mode not in [
                    "direct", "threshold", "threshold-one-hot"
            ]:
                raise ValueError(
                    "classification_handling_mode must be one of 'direct', 'threshold', 'threshold_one_hot'"
                )

        self.mode = mode
        self.n_tasks = n_tasks
        self.classification_handling_mode = classification_handling_mode
        self.threshold_value = threshold_value

    def compute_metric(self,
                       y_true: ArrayLike,
                       y_pred: ArrayLike,
                       w: Optional[ArrayLike] = None,
                       n_tasks: Optional[int] = None,
                       n_classes: int = 2,
                       per_task_metrics: bool = False,
                       use_sample_weights: bool = False,
                       **kwargs) -> Any:
        """Compute a performance metric for each task.

        Parameters
        ----------
        y_true: ArrayLike
            An ArrayLike containing true values for each task. Must be of shape
            `(N,)` or `(N, n_tasks)` or `(N, n_tasks, n_classes)` if a
            classification metric. If of shape `(N, n_tasks)` values can either be
            class-labels or probabilities of the positive class for binary
            classification problems. If a regression problem, must be of shape
            `(N,)` or `(N, n_tasks)` or `(N, n_tasks, 1)` if a regression metric.
        y_pred: ArrayLike
            An ArrayLike containing predicted values for each task. Must be
            of shape `(N, n_tasks, n_classes)` if a classification metric,
            else must be of shape `(N, n_tasks)` if a regression metric.
        w: ArrayLike, default None
            An ArrayLike containing weights for each datapoint. If
            specified,  must be of shape `(N, n_tasks)`.
        n_tasks: int, default None
            The number of tasks this class is expected to handle.
        n_classes: int, default 2
            Number of classes in data for classification tasks.
        per_task_metrics: bool, default False
            If true, return computed metric for each task on multitask dataset.
        use_sample_weights: bool, default False
            If set, use per-sample weights `w`.
        kwargs: dict
            Will be passed on to self.metric

        Returns
        -------
        np.ndarray
            A numpy array containing metric values for each task.
        """
        # Attempt some limited shape imputation to find n_tasks
        y_true_arr = np.asarray(y_true)
        y_pred_arr = np.asarray(y_pred)
        if n_tasks is None:
            if self.n_tasks is None and isinstance(y_true_arr, np.ndarray):
                if len(y_true_arr.shape) == 1:
                    n_tasks = 1
                elif len(y_true_arr.shape) >= 2:
                    n_tasks = y_true_arr.shape[1]
            else:
                n_tasks = self.n_tasks
        # check whether n_tasks is int or not
        # This is because `normalize_weight_shape` require int value.
        assert isinstance(n_tasks, int)

        y_true_arr = normalize_labels_shape(y_true_arr,
                                            mode=self.mode,
                                            n_tasks=n_tasks,
                                            n_classes=n_classes)
        y_pred_arr = normalize_prediction_shape(y_pred_arr,
                                                mode=self.mode,
                                                n_tasks=n_tasks,
                                                n_classes=n_classes)
        if self.mode == "classification":
            y_true_arr = handle_classification_mode(
                y_true_arr, self.classification_handling_mode,
                self.threshold_value)
            y_pred_arr = handle_classification_mode(
                y_pred_arr, self.classification_handling_mode,
                self.threshold_value)
        n_samples = y_true_arr.shape[0]
        w = normalize_weight_shape(None if w is None else np.asarray(w),
                                   n_samples, n_tasks)
        computed_metrics = []
        for task in range(n_tasks):
            y_task = y_true_arr[:, task]
            y_pred_arr_task = y_pred_arr[:, task]
            w_task = w[:, task]

            metric_value = self.compute_singletask_metric(
                y_task,
                y_pred_arr_task,
                w_task,
                use_sample_weights=use_sample_weights,
                **kwargs)
            computed_metrics.append(metric_value)
        logger.info("computed_metrics: %s" % str(computed_metrics))
        if n_tasks == 1:
            # FIXME: Incompatible types in assignment
            computed_metrics = computed_metrics[0]  # type: ignore

        if not per_task_metrics:
            return self.task_averager(computed_metrics)
        else:
            return self.task_averager(computed_metrics), computed_metrics

    def compute_singletask_metric(self,
                                  y_true: ArrayLike,
                                  y_pred: ArrayLike,
                                  w: Optional[ArrayLike] = None,
                                  n_samples: Optional[int] = None,
                                  use_sample_weights: bool = False,
                                  **kwargs) -> float:
        """Compute a metric value.

        Parameters
        ----------
        y_true: ArrayLike
            True values array. This array must be of shape `(N,
            n_classes)` if classification and `(N,)` if regression.
        y_pred: ArrayLike
            Predictions array. This array must be of shape `(N, n_classes)`
            if classification and `(N,)` if regression.
        w: ArrayLike, default None
            Sample weight array. This array must be of shape `(N,)`
        n_samples: int, default None (DEPRECATED)
            The number of samples in the dataset. This is `N`. This argument is
            ignored.
        use_sample_weights: bool, default False
            If set, use per-sample weights `w`.
        kwargs: dict
            Will be passed on to self.metric

        Returns
        -------
        metric_value: float
            The computed value of the metric.
        """
        if n_samples is not None:
            logger.warning(
                "n_samples is a deprecated argument which is ignored.")
        # Attempt to convert both into the same type
        y_true_arr = np.asarray(y_true)
        y_pred_arr = np.asarray(y_pred)
        if self.mode == "regression":
            if len(y_true_arr.shape) != 1 or len(
                    y_pred_arr.shape
            ) != 1 or y_true_arr.shape != y_pred_arr.shape:
                raise ValueError(
                    "For regression metrics, y_true and y_pred must both be of shape (N,)"
                )
        elif self.mode == "classification":
            pass
            # if len(y_true.shape) != 2 or len(y_pred.shape) != 2 or y_true.shape != y_pred.shape:
            # raise ValueError("For classification metrics, y_true and y_pred must both be of shape (N, n_classes)")
        else:
            raise ValueError(
                "Only classification and regression are supported for metrics calculations."
            )
        if use_sample_weights:
            metric_value = self.metric(y_true_arr,
                                       y_pred_arr,
                                       sample_weight=w,
                                       **kwargs)
        else:
            metric_value = self.metric(y_true_arr, y_pred_arr, **kwargs)
        return metric_value



import unittest
import os
import platform
from deepchem.utils import sequence_utils as seq_utils

IS_WINDOWS = platform.system() == 'Windows'


@unittest.skipIf(IS_WINDOWS,
                 "Skip test on Windows")  # hhsuite does not run on windows
class TestSeq(unittest.TestCase):
    """
    Tests sequence handling utilities.
    """

    def setUp(self):
        current_dir = os.path.dirname(os.path.realpath(__file__))
        self.dataset_file = os.path.join(current_dir, 'assets/example.fasta')
        self.database_name = 'example_db'
        self.data_dir = os.path.join(current_dir, 'assets')

    def test_hhsearch(self):
        seq_utils.hhsearch(self.dataset_file,
                           database=self.database_name,
                           data_dir=self.data_dir)
        results_file = os.path.join(self.data_dir, 'results.a3m')
        hhr_file = os.path.join(self.data_dir, 'example.hhr')
        with open(results_file, 'r') as f:
            resultsline = next(f)
        with open(hhr_file, 'r') as g:
            hhrline = next(g)

        assert hhrline[0:5] == 'Query'
        assert resultsline[0:5] == '>seq0'
        os.remove(results_file)
        os.remove(hhr_file)

    def test_hhblits(self):
        seq_utils.hhsearch(self.dataset_file,
                           database=self.database_name,
                           data_dir=self.data_dir)
        results_file = os.path.join(self.data_dir, 'results.a3m')
        hhr_file = os.path.join(self.data_dir, 'example.hhr')
        with open(results_file, 'r') as f:
            resultsline = next(f)
        with open(hhr_file, 'r') as g:
            hhrline = next(g)

        assert hhrline[0:5] == 'Query'
        assert resultsline[0:5] == '>seq0'
        os.remove(results_file)
        os.remove(hhr_file)

    def test_MSA_to_dataset(self):
        seq_utils.hhsearch(self.dataset_file,
                           database=self.database_name,
                           data_dir=self.data_dir)
        results_file = os.path.join(self.data_dir, 'results.a3m')
        msa_path = results_file
        dataset = seq_utils.MSA_to_dataset(msa_path)
        print(dataset.ids[0])
        print(dataset.X)
        assert dataset.ids[0] == 'seq0'
        assert dataset.ids[1] == 'seq1'
        bool_arr = dataset.X[0] == ['X', 'Y']
        assert bool_arr.all()
        os.remove(results_file)



from typing import List, Union, Optional, Tuple, Dict
import warnings
import torch
from deepchem.utils.dft_utils import BaseHamilton, BaseSystem, BaseGrid, \
    CGTOBasis, AtomCGTOBasis, SpinParam, ZType, BasisInpType, AtomZsType, \
    AtomPosType, loadbasis, parse_moldesc, get_predefined_grid, HamiltonCGTO, \
    is_z_float, DensityFitInfo
from deepchem.utils.periodic_table_utils import get_atomz, get_atom_mass
from deepchem.utils import occnumber, safe_cdist
from deepchem.utils.cache_utils import Cache
import logging

logger = logging.getLogger(__name__)


class Mol(BaseSystem):
    """
    Describe the system of an isolated molecule

    Examples
    --------
    >>> from deepchem.utils.dft_utils import Mol
    >>> mol = Mol("H 1 0 0; H -1 0 0", "sto-3g", spin=1)
    >>> mol.setup_grid()
    >>> mol.get_orbweight()
    tensor([1.5000, 0.5000], dtype=torch.float64)

    """

    def __init__(
            self,
            moldesc: Union[str, Tuple[AtomZsType, AtomPosType]],
            basis: BasisInpType,
            *,
            grid: Union[int, str] = "sg3",
            spin: Optional[ZType] = None,
            charge: ZType = 0,
            orb_weights: Optional[SpinParam[torch.Tensor]] = None,
            efield: Union[torch.Tensor, Tuple[torch.Tensor, ...], None] = None,
            dtype: torch.dtype = torch.float64,
            device: torch.device = torch.device('cpu'),
    ):
        """Initialize the molecule system.

        Parameters
        ----------
        moldesc: str or 2-elements tuple
            Description of the molecule system.
            If string, it can be described like ``"H 1 0 0; H -1 0 0"``.
            If tuple, the first element of the tuple is the Z number of the atoms while
            the second element is the position of the atoms: ``(atomzs, atomposs)``.
        basis: str, CGTOBasis, list of str, or CGTOBasis
            The string describing the gto basis. If it is a list, then it must have
            the same length as the number of atoms.
        grid: int
            Describe the grid.
            If it is an integer, then it uses the default grid with specified level
            of accuracy.
        spin: int, float, torch.Tensor, or None
            The difference between spin-up and spin-down electrons.
            It must be an integer or ``None``.
            If ``None``, then it is ``num_electrons % 2``.
            For floating point atomzs and/or charge, the ``spin`` must be specified.
        charge: int, float, or torch.Tensor
            The charge of the molecule.
        orb_weights: SpinParam[torch.Tensor] or None
            Specifiying the orbital occupancy (or weights) directly. If specified,
            ``spin`` and ``charge`` arguments are ignored.
        efield: tensor, tuple of tensor, or None
            Uniform electric field of the system. If a tensor, then it is assumed
            to be a constant electric field with the energy is
            calculated based on potential at ``(0, 0, 0)`` is ``0``.
            If a tuple of tensor, then the first element will have a shape of ``(ndim,)``
            representing the constant electric field, second element is the gradient
            of electric field with the last dimension is the direction of the electric
            field, third element is the gradgrad of electric field, etc.
            If ``None``, then the electric field is assumed to be ``0``.
        dtype: torch.dtype
            The data type of tensors in this class.
        device: torch.device
            The device on which the tensors in this class are stored.

        """
        self._dtype = dtype
        self._device = device
        self._grid_inp = grid
        self._basis_inp = basis
        self._grid: Optional[BaseGrid] = None

        # make efield a tuple
        self._efield = _normalize_efield(efield)
        self._preproc_efield = _preprocess_efield(self._efield)

        # initialize cache
        self._cache = Cache()

        # get the AtomCGTOBasis & the hamiltonian
        # atomzs: (natoms,) dtype: torch.int or dtype for floating point
        # atompos: (natoms, ndim)
        atomzs, atompos = parse_moldesc(moldesc, dtype=dtype, device=device)
        atomzs_int = torch.round(atomzs).to(
            torch.int) if atomzs.is_floating_point() else atomzs
        allbases = _parse_basis(atomzs_int, basis)  # list of list of CGTOBasis
        atombases = [
            AtomCGTOBasis(atomz=atz, bases=bas, pos=atpos)
            for (atz, bas, atpos) in zip(atomzs, allbases, atompos)
        ]
        self._atombases = atombases
        self._hamilton = HamiltonCGTO(atombases,
                                      efield=self._preproc_efield,
                                      cache=self._cache.add_prefix("hamilton"))
        self._atompos = atompos  # (natoms, ndim)
        self._atomzs = atomzs  # (natoms,) int-type or dtype if floating point
        self._atomzs_int = atomzs_int  # (natoms,) int-type rounded from atomzs
        nelecs_tot: torch.Tensor = torch.sum(atomzs)

        # orb_weights is not specified, so determine it from spin and charge
        if orb_weights is None:
            # get the number of electrons and spin
            nelecs, spin, frac_mode = _get_nelecs_spin(nelecs_tot, spin, charge)
            _orb_weights, _orb_weights_u, _orb_weights_d = _get_orb_weights(
                nelecs, spin, frac_mode, dtype, device)

            # save the system's properties
            self._spin = spin
            self._charge = charge
            self._numel = nelecs
            self._orb_weights = _orb_weights
            self._orb_weights_u = _orb_weights_u
            self._orb_weights_d = _orb_weights_d

        # orb_weights is specified, so calculate the spin and charge from it
        else:
            if not isinstance(orb_weights, SpinParam):
                raise TypeError(
                    "Specifying orb_weights must be in SpinParam type")
            assert orb_weights.u.ndim == 1
            assert orb_weights.d.ndim == 1
            assert len(orb_weights.u) == len(orb_weights.d)

            # check if it is decreasing
            orb_u_dec = torch.all(
                orb_weights.u[:-1] - orb_weights.u[1:] > -1e-4)
            orb_d_dec = torch.all(
                orb_weights.d[:-1] - orb_weights.d[1:] > -1e-4)
            if not (orb_u_dec and orb_d_dec):
                # if not decreasing, the variational might give the wrong results
                warnings.warn(
                    "The orbitals should be ordered in a non-increasing manner. "
                    "Otherwise, some calculations might be wrong.")

            utot = orb_weights.u.sum()
            dtot = orb_weights.d.sum()
            self._numel = utot + dtot
            self._spin = utot - dtot
            self._charge = nelecs_tot - self._numel

            self._orb_weights_u = orb_weights.u
            self._orb_weights_d = orb_weights.d
            self._orb_weights = orb_weights.u + orb_weights.d

    def densityfit(self,
                   method: Optional[str] = None,
                   auxbasis: Optional[BasisInpType] = None) -> BaseSystem:
        """
        Indicate that the system's Hamiltonian uses density fit for its integral.

        Parameters
        ----------
        method: Optional[str]
            Density fitting method. Available methods in this class are:

            * ``"coulomb"``: Minimizing the Coulomb inner product, i.e. ``min <p-p_fit|r_12|p-p_fit>``
              Ref: Eichkorn, et al. Chem. Phys. Lett. 240 (1995) 283-290.
              (default)
            * ``"overlap"``: Minimizing the overlap inner product, i.e. min <p-p_fit|p-p_fit>

        auxbasis: Optional[BasisInpType]
            Auxiliary basis for the density fit. If not specified, then it uses
            ``"cc-pvtz-jkfit"``.

        Returns
        -------
        BaseSystem
            The system with the density fit Hamiltonian.

        """
        if method is None:
            method = "coulomb"
        if auxbasis is None:
            # TODO: choose the auxbasis properly
            auxbasis = "cc-pvtz-jkfit"

        # get the auxiliary basis
        assert auxbasis is not None
        auxbasis_lst = _parse_basis(self._atomzs_int, auxbasis)
        atomauxbases = [
            AtomCGTOBasis(atomz=atz, bases=bas, pos=atpos)
            for (atz, bas,
                 atpos) in zip(self._atomzs, auxbasis_lst, self._atompos)
        ]

        # change the hamiltonian to have density fit
        df = DensityFitInfo(method=method, auxbasis=atomauxbases)
        self._hamilton = HamiltonCGTO(self._atombases,
                                      df=df,
                                      efield=self._preproc_efield,
                                      cache=self._cache.add_prefix("hamilton"))
        return self

    def get_hamiltonian(self) -> BaseHamilton:
        """
        Hamiltonian that corresponds to the system.

        Returns
        -------
        BaseHamilton
            The Hamiltonian of the system.

        """
        return self._hamilton

    def set_cache(self,
                  fname: str,
                  paramnames: Optional[List[str]] = None) -> BaseSystem:
        """
        Setup the cache of some parameters specified by `paramnames` to be read/written
        on a file.
        If the file exists, then the parameters will not be recomputed, but just
        loaded from the cache instead.
        Cache is usually used for repeated calculations where the cached parameters
        are not changed (e.g. running multiple systems with slightly different environment.)

        Parameters
        ----------
        fname: str
            The file to store the cache.
        paramnames: list of str or None
            List of parameter names to be read/write from the cache.

        Returns
        -------
        BaseSystem
            The system with the cache setup.

        """
        all_paramnames = self._cache.get_cacheable_params()
        if paramnames is not None:
            # check the paramnames
            for pname in paramnames:
                if pname not in all_paramnames:
                    msg = "Parameter %s is not cache-able. Cache-able parameters are %s" % \
                        (pname, all_paramnames)
                    raise ValueError(msg)

        self._cache.set(fname, paramnames)

        return self

    def get_orbweight(
        self,
        polarized: bool = False
    ) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
        """Get the orbital weights of the system.

        Parameters
        ----------
        polarized: bool
            If True, then return the polarized orbital weights.

        Returns
        -------
        torch.Tensor or SpinParam[torch.Tensor]
            The orbital weights of the system.
            If polarized is True, then it returns the polarized orbital weights.

        """
        if not polarized:
            return self._orb_weights
        else:
            return SpinParam(u=self._orb_weights_u, d=self._orb_weights_d)

    def get_nuclei_energy(self) -> torch.Tensor:
        """Get the nuclei-nuclei energy of the system.

        Returns
        -------
        torch.Tensor
            Nuclei-nuclei energy of the system.

        """
        # atomzs: (natoms,)
        # atompos: (natoms, ndim)

        # r12: (natoms, natom)
        r12 = safe_cdist(self._atompos,
                         self._atompos,
                         add_diag_eps=True,
                         diag_inf=True)
        z12 = self._atomzs.unsqueeze(-2) * self._atomzs.unsqueeze(
            -1)  # (natoms, natoms)
        q_by_r = z12 / r12
        return q_by_r.sum() * 0.5

    def setup_grid(self) -> None:
        """Setup the integration grid for the system."""
        logger.info("Constructing the integration grid")
        self._grid = get_predefined_grid(self._grid_inp,
                                         self._atomzs_int,
                                         self._atompos,
                                         dtype=self._dtype,
                                         device=self._device)
        logger.info("Constructing the integration grid: done")

        # #        0,  1,  2,  3,  4,  5
        # nr   = [20, 40, 60, 75, 100, 125][grid_inp]
        # prec = [13, 17, 21, 29, 41, 59][grid_inp]
        # radgrid = RadialGrid(nr, "chebyshev", "logm3",
        #                      dtype=self._dtype, device=self._device)
        # sphgrid = LebedevGrid(radgrid, prec=prec)
        #
        # natoms = self._atompos.shape[-2]
        # sphgrids = [sphgrid for _ in range(natoms)]
        # self._grid = BeckeGrid(sphgrids, self._atompos)

    def get_grid(self) -> BaseGrid:
        """Get the integration grid of the system.

        Returns
        -------
        BaseGrid
            The integration grid of the system.

        """
        if self._grid is None:
            raise RuntimeError(
                "Please run mol.setup_grid() first before calling get_grid()")
        return self._grid

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """Get the parameter names of the system.

        Parameters
        ----------
        methodname: str
            Method name to get the parameter names.
        prefix: str
            Prefix to be added to the parameter names.

        Returns
        -------
        List[str]
            List of parameter names.

        """
        if methodname == "get_nuclei_energy":
            params = [prefix + "_atompos"]
            if torch.is_floating_point(self._atomzs):
                params += [prefix + "_atomzs"]
            return params
        else:
            raise KeyError("Unknown methodname: %s" % methodname)

    # properties
    @property
    def atompos(self) -> torch.Tensor:
        """Get the atomic positions of the system.

        Returns
        -------
        torch.Tensor
            Atomic positions of the system.

        """
        return self._atompos

    @property
    def atomzs(self) -> torch.Tensor:
        """Get the atomic numbers of the system.

        Returns
        -------
        torch.Tensor
            Atomic numbers of the system.

        """
        return self._atomzs

    @property
    def atommasses(self) -> torch.Tensor:
        """returns the atomic mass (only for non-isotope for now)

        Returns
        -------
        torch.Tensor
            Atomic masses of the system.

        """
        if torch.is_floating_point(self._atomzs):
            raise RuntimeError(
                "Atom masses are not available for floating point Z")
        return torch.tensor(
            [get_atom_mass(int(atomz)) for atomz in self._atomzs],
            dtype=self._dtype,
            device=self._device)

    @property
    def spin(self) -> ZType:
        """Get the spin of the system.

        Returns
        -------
        ZType
            Spin of the system.

        """
        return self._spin

    @property
    def charge(self) -> ZType:
        """Get the charge of the system.

        Returns
        -------
        ZType
            Charge of the system.

        """
        return self._charge

    @property
    def numel(self) -> ZType:
        """Get the number of electrons of the system.

        Returns
        -------
        ZType
            Number of electrons of the system.

        """
        return self._numel

    @property
    def efield(self) -> Optional[Tuple[torch.Tensor, ...]]:
        """Get the electric field of the system.

        Returns
        -------
        Optional[Tuple[torch.Tensor, ...]]
            Electric field of the system.

        """
        return self._efield


def _parse_basis(atomzs: torch.Tensor,
                 basis: BasisInpType) -> List[List[CGTOBasis]]:
    """Returns the list of cgto basis for every atoms.

    Examples
    --------
    >>> _parse_basis(torch.tensor([1, 1]), "sto-3g")[0][0].alphas
    tensor([3.4253, 0.6239, 0.1689], dtype=torch.float64)

    Parameters
    ----------
    atomzs: torch.Tensor
        Atomic numbers of the atoms.
    basis: BasisInpType
        The basis input.

    Returns
    -------
    List[List[CGTOBasis]]
        List of CGTOBasis for every atoms.

    """
    natoms = len(atomzs)

    if isinstance(basis, str):
        return [loadbasis("%d:%s" % (atomz, basis)) for atomz in atomzs]

    elif isinstance(basis, dict):
        # convert the basis key into atomz
        basis_int: Dict[int, List[CGTOBasis]] = {}
        for k, v in basis.items():
            atz = int(get_atomz(k))
            bas = v if isinstance(v, list) else loadbasis("%d:%s" % (atz, v))
            basis_int[atz] = bas
        return [basis_int[int(atomz)] for atomz in atomzs]

    else:  # basis is a list
        assert len(atomzs) == len(basis)

        # TODO: safely remove "type: ignore" in this block
        assert len(basis) > 0

        # list of cgtobasis
        if isinstance(basis[0], CGTOBasis):
            return [basis for _ in range(natoms)]  # type: ignore

        # list of str
        elif isinstance(basis[0], str):
            return [
                loadbasis("%d:%s" % (atz, b))
                for (atz, b) in zip(atomzs, basis)
            ]  # type: ignore

        # list of list of cgto basis
        else:
            return basis  # type: ignore


def _get_nelecs_spin(nelecs_tot: torch.Tensor, spin: Optional[ZType],
                     charge: ZType) -> Tuple[torch.Tensor, ZType, bool]:
    """get the number of electrons and spins

    Examples
    --------
    >>> _get_nelecs_spin(torch.tensor(2), None, 0)
    (tensor(2), tensor(0), False)

    Parameters
    ----------
    nelecs_tot: torch.Tensor
        Total number of electrons.
    spin: int, float, torch.Tensor, or None
        Spin of the system.
    charge: int, float, or torch.Tensor
        Charge of the system.

    Returns
    -------
    Tuple[torch.Tensor, ZType, bool]
        Number of electrons, spin, and whether the system is in fractional mode.

    """

    # a boolean to indicate if working in a fractional mode
    frac_mode = nelecs_tot.is_floating_point() or is_z_float(charge) or \
        (spin is not None and is_z_float(spin))

    assert nelecs_tot >= charge, \
        "Only %f electrons, but needs %f charge" % (nelecs_tot.item(), charge)
    nelecs: torch.Tensor = nelecs_tot - charge

    # if spin is not given, then set it as the remainder if nelecs is an integer
    if spin is None:
        assert not frac_mode, \
            "Fraction case requires the spin argument to be specified"
        spin = nelecs % 2
    else:
        assert spin >= 0
        if not frac_mode:
            # only check if the calculation is not in fraction mode,
            # for fractional mode, unmatched spin is acceptable
            assert (nelecs - spin) % 2 == 0, \
                "Spin %d is not suited for %d electrons" % (spin, nelecs)
    return nelecs, spin, frac_mode


def _get_orb_weights(nelecs: torch.Tensor, spin: ZType, frac_mode: bool,
                     dtype: torch.dtype, device: torch.device) \
        -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """returns the orbital weights given the electronic information
    (total orbital weights, spin-up orb weights, spin-down orb weights)

    Examples
    --------
    >>> _get_orb_weights(torch.tensor(2), 1, False, torch.float64, torch.device('cpu'))
    (tensor([1.], dtype=torch.float64), tensor([1.], dtype=torch.float64), tensor([0.], dtype=torch.float64))

    Parameters
    ----------
    nelecs: torch.Tensor
        Number of electrons.
    spin: int
        Spin of the system.
    frac_mode: bool
        Whether the system is in fractional mode.
    dtype: torch.dtype
        Data type of the tensors.
    device: torch.device
        Device of the tensors.

    Returns
    -------
    Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
        Total orbital weights, spin-up orbital weights, and spin-down orbital weights.

    """

    # calculate the orbital weights
    nspin_dn: torch.Tensor = (nelecs - spin) * 0.5 if frac_mode else \
        torch.div(nelecs - spin, 2, rounding_mode="floor")
    nspin_up: torch.Tensor = nspin_dn + spin

    # total orbital weights
    _orb_weights_u = occnumber(nspin_up, dtype=dtype, device=device)
    _orb_weights_d = occnumber(nspin_dn,
                               n=len(_orb_weights_u),
                               dtype=dtype,
                               device=device)
    _orb_weights = _orb_weights_u + _orb_weights_d

    # get the polarized orbital weights
    if nspin_dn > 0:
        _orb_weights_d = occnumber(nspin_dn, dtype=dtype, device=device)
    else:
        _orb_weights_d = occnumber(0, n=1, dtype=dtype, device=device)

    return _orb_weights, _orb_weights_u, _orb_weights_d


def _normalize_efield(efield: Union[torch.Tensor, Tuple[torch.Tensor, ...], None]) \
        -> Optional[Tuple[torch.Tensor, ...]]:
    """making efield a tuple or None

    Examples
    --------
    >>> _normalize_efield(torch.tensor([1, 2, 3]))
    (tensor([1, 2, 3]),)

    Parameters
    ----------
    efield: torch.Tensor, tuple of torch.Tensor, or None
        Electric field of the system.

    Returns
    -------
    Optional[Tuple[torch.Tensor, ...]]
        Electric field of the system.

    """

    if isinstance(efield, torch.Tensor):
        efs: Optional[Tuple[torch.Tensor, ...]] = (efield,)
    else:
        efs = efield

    return efs


def _preprocess_efield(
    efs: Optional[Tuple[torch.Tensor,
                        ...]]) -> Optional[Tuple[torch.Tensor, ...]]:
    """preprocess the efield tuple such that the energy is just a simple tensor
    product with the integrals

    Examples
    --------
    >>> _preprocess_efield((torch.tensor([1, 2, 3]),))
    (tensor([1, 2, 3]),)

    Parameters
    ----------
    efs: tuple of torch.Tensor or None
        Electric field of the system.

    Returns
    -------
    tuple of torch.Tensor or None
        Preprocessed electric field of the system

    """

    if efs is None:
        return efs

    assert isinstance(efs, tuple)
    res_list: List[torch.Tensor] = []
    for i in range(len(efs)):
        efi = efs[i]
        numel = 3**(i + 1)
        assert efi.numel() == numel, \
            f"The {i}-th tuple element of efield must have {numel} elements"

        res_list.append(efi.reshape(-1))

    return tuple(res_list)



import torch
import torch.nn as nn
import torch.nn.functional as F
import logging
from typing import List, Tuple, Callable, Literal, Union
from typing import Sequence as SequenceCollection
from deepchem.utils.typing import OneOrMany, ActivationFn
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models import losses
from deepchem.metrics import to_one_hot
import datetime

logger = logging.getLogger(__name__)


class RobustMultitask(nn.Module):
    """Implements a neural network for robust multitasking.

    The key idea of this model is to have bypass layers that feed
    directly from features to task output. This might provide some
    flexibility toroute around challenges in multitasking with
    destructive interference.

    References
    ----------
    This technique was introduced in [1]_

    .. [1] Ramsundar, Bharath, et al. "Is multitask deep learning practical for pharma?." Journal of chemical information and modeling 57.8 (2017): 2068-2076.

    """

    def __init__(self,
                 n_tasks: int,
                 n_features: int,
                 layer_sizes: SequenceCollection[int] = [1000],
                 mode: Literal['regression', 'classification'] = 'regression',
                 weight_init_stddevs: OneOrMany[float] = 0.02,
                 bias_init_consts: OneOrMany[float] = 1.0,
                 weight_decay_penalty=0.0,
                 weight_decay_penalty_type="l2",
                 activation_fns: OneOrMany[ActivationFn] = nn.ReLU(),
                 dropouts: OneOrMany[float] = 0.5,
                 n_classes: int = 2,
                 bypass_layer_sizes: SequenceCollection[int] = [100],
                 bypass_weight_init_stddevs: OneOrMany[float] = [.02],
                 bypass_bias_init_consts: OneOrMany[float] = [1.0],
                 bypass_dropouts: OneOrMany[float] = [0.5],
                 **kwargs):

        self.n_tasks: int = n_tasks
        self.n_features: int = n_features
        self.n_classes: int = n_classes
        self.mode: Literal['regression', 'classification'] = mode
        self.layer_sizes: SequenceCollection[int] = layer_sizes
        self.bypass_layer_sizes: SequenceCollection[int] = ([
            bypass_layer_sizes
        ] if isinstance(bypass_layer_sizes, int) else bypass_layer_sizes)
        self.weight_decay_penalty: float = weight_decay_penalty
        self.weight_decay_penalty_type: Literal[
            'l1', 'l2'] = weight_decay_penalty_type
        n_layers: int = len(layer_sizes)
        n_bypass_layers: int = len(bypass_layer_sizes)

        if not isinstance(weight_init_stddevs, SequenceCollection):
            weight_init_stddevs = [weight_init_stddevs] * n_layers
        if not isinstance(bias_init_consts, SequenceCollection):
            bias_init_consts = [bias_init_consts] * n_layers
        if not isinstance(dropouts, SequenceCollection):
            dropouts = [dropouts] * n_layers
        if not isinstance(bypass_weight_init_stddevs, SequenceCollection):
            bypass_weight_init_stddevs = [bypass_weight_init_stddevs
                                         ] * n_bypass_layers
        if not isinstance(bypass_bias_init_consts, SequenceCollection):
            bypass_bias_init_consts = [bypass_bias_init_consts
                                      ] * n_bypass_layers
        if not isinstance(bypass_dropouts, SequenceCollection):
            bypass_dropouts = [bypass_dropouts] * n_bypass_layers
        if isinstance(
                activation_fns,
                str) or not isinstance(activation_fns, SequenceCollection):
            activation_fns = [activation_fns] * n_layers

        self.activation_fns: SequenceCollection[ActivationFn] = [
            self._get_activation_class(f) for f in activation_fns
        ]
        self.weight_init_stddevs: SequenceCollection[
            float] = weight_init_stddevs
        self.bias_init_consts: SequenceCollection[float] = bias_init_consts
        self.dropouts: SequenceCollection[float] = dropouts
        self.bypass_activation_fns: SequenceCollection[ActivationFn] = [
            self.activation_fns[0]
        ] * n_bypass_layers
        self.bypass_weight_init_stddevs: SequenceCollection[
            float] = bypass_weight_init_stddevs
        self.bypass_bias_init_consts: SequenceCollection[
            float] = bypass_bias_init_consts

        super(RobustMultitask, self).__init__()

        # Add shared layers
        self.shared_layers: nn.Sequential = self._build_layers(
            n_features, layer_sizes, self.activation_fns, dropouts,
            self.weight_init_stddevs, self.bias_init_consts)

        # Add task-specific bypass layers
        self.bypass_layers: nn.ModuleList = nn.ModuleList([
            self._build_layers(n_features, bypass_layer_sizes,
                               self.bypass_activation_fns, bypass_dropouts,
                               self.bypass_weight_init_stddevs,
                               self.bypass_bias_init_consts)
            for _ in range(n_tasks)
        ])

        # Output layers for each task
        self.output_layers: nn.ModuleList = nn.ModuleList([
            nn.Linear(layer_sizes[-1] + bypass_layer_sizes[-1], n_classes)
            for _ in range(n_tasks)
        ])

    def _build_layers(self, input_size, layer_sizes, activation_fns, dropouts,
                      weight_init_stddevs, bias_init_consts):
        """Helper function to build layers with activations and dropout"""

        prev_size = input_size
        layers = []

        for i, size in enumerate(layer_sizes):
            layer = nn.Linear(prev_size, size)
            try:
                nn.init.trunc_normal_(layer.weight, std=weight_init_stddevs[i])
            except IndexError:
                logger.warning(
                    "Warning: Wrong number of weight_init_stddevs specified. When passing weight_init_stddevs as a list, the length of the list should be equal to the number of layers."
                )
                logger.warning(
                    "Using default weight initialization: truncated normal with std=0.02"
                )
                nn.init.trunc_normal_(layer.weight, std=0.02)
            try:
                nn.init.constant_(layer.bias, bias_init_consts[i])
            except IndexError:
                logger.warning(
                    "Warning: Wrong number of bias_init_consts specified. When passing bias_init_consts as a list, the length of the list should be equal to the number of layers."
                )
                logger.warning(
                    "Using default bias initialization: constant=1.0")
                nn.init.constant_(layer.bias, 1.0)

            layers.append(layer)

            try:
                layers.append(activation_fns[i])
            except IndexError:
                logger.warning(
                    "Warning: Mismatch in number of activation functions and layers detected."
                )
                pass
            try:
                layers.append(nn.Dropout(dropouts[i]))
            except IndexError:
                logger.warning(
                    "Warning: Mismatch in number of dropouts and layers detected."
                )
                pass

            prev_size = size

        sequential = nn.Sequential(*layers)
        return sequential

    def forward(
        self, x: torch.Tensor
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """Forward pass through the network.

        Parameters
        ----------
        x: torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            The Model output tensor of shape (batch_size, n_tasks, n_outputs).

        * When self.mode = `regression`,
            It consists of the output of each task.
        * When self.mode = `classification`,
            It consists of the probability of each class for each task.

        torch.Tensor, optional
            This is only returned when self.mode = `classification`, the output consists of the
            logits for classes before softmax.
        """
        task_outputs: List[torch.Tensor] = []

        # Shared layers
        shared_output = self.shared_layers(x)

        # Bypass layers
        for task in range(self.n_tasks):
            bypass_output = self.bypass_layers[task](x)

            # Concatenating outputs of shared layers and each task's bypass layers
            combined_output = torch.cat([shared_output, bypass_output], dim=1)

            # Task specific output layer
            task_output = self.output_layers[task](combined_output)
            task_outputs.append(task_output)

        output = torch.stack(task_outputs, dim=1)

        if self.mode == 'classification':
            if self.n_tasks == 1:
                logits = output.view(-1, self.n_classes)
                softmax_dim = 1
            else:
                logits = output.view(-1, self.n_tasks, self.n_classes)
                softmax_dim = 2
            proba = F.softmax(logits, dim=softmax_dim)
            return proba, logits
        else:
            return output.squeeze(-2)

    def regularization_loss(self):
        """Compute the regularization loss for the model."""
        reg_loss = 0.0

        for param in self.parameters():
            if self.weight_decay_penalty_type == "l1":
                reg_loss += torch.sum(torch.abs(param))
            elif self.weight_decay_penalty_type == "l2":
                reg_loss += torch.sum(torch.pow(param, 2))

        return self.weight_decay_penalty * reg_loss

    def _get_activation_class(self, activation_name: ActivationFn) -> Callable:
        """Get the activation class from the name of the activation function.

        Parameters
        ----------
        activation_name: str
            The name of the activation function.

        Returns
        -------
        Callable
            The activation function class.
        """
        if isinstance(activation_name, str):
            return getattr(nn, activation_name)
        elif isinstance(activation_name, nn.Module):
            return activation_name
        else:
            raise ValueError(
                f"Invalid activation function: {activation_name}. Only activations of type torch.nn.Module (torch.nn.functional activations are not supported yet!!)"
            )


class RobustMultitaskClassifier(TorchModel):
    """
    Implements a neural network for robust multitasking.
    The key idea of this model is to have bypass layers that feed
    directly from features to task output. This might provide some
    flexibility toroute around challenges in multitasking with
    destructive interference.
    References
    ----------
    This technique was introduced in [1]_
    .. [1] Ramsundar, Bharath, et al. "Is multitask deep learning practical for pharma?." Journal of chemical information and modeling 57.8 (2017): 2068-2076.
    """

    def __init__(self,
                 n_tasks: int,
                 n_features: int,
                 layer_sizes: SequenceCollection[int] = [1000],
                 weight_init_stddevs: OneOrMany[float] = 0.02,
                 bias_init_consts: OneOrMany[float] = 1.0,
                 weight_decay_penalty: float = 0.0,
                 weight_decay_penalty_type: Literal['l1', 'l2'] = "l2",
                 dropouts: OneOrMany[float] = 0.5,
                 activation_fns: OneOrMany[ActivationFn] = nn.ReLU(),
                 n_classes: int = 2,
                 bypass_layer_sizes: SequenceCollection[int] = [100],
                 bypass_weight_init_stddevs: OneOrMany[float] = [0.02],
                 bypass_bias_init_consts: OneOrMany[float] = [1.0],
                 bypass_dropouts: OneOrMany[float] = [0.5],
                 **kwargs):
        """
        Parameters
        ----------
        n_tasks: int
            number of tasks
        n_features: int
            number of features
        layer_sizes: list
            the size of each dense layer in the network.  The length of this list determines the number of layers.
        weight_init_stddevs: list or float
            the standard deviation of the distribution to use for weight initialization of each layer.  The length
            of this list should equal len(layer_sizes).  Alternatively this may be a single value instead of a list,
            in which case the same value is used for every layer.
        bias_init_consts: list or loat
            the value to initialize the biases in each layer to.  The length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        weight_decay_penalty: float
            the magnitude of the weight decay penalty to use
        weight_decay_penalty_type: str
            the type of penalty to use for weight decay, either 'l1' or 'l2'
        dropouts: list or float
            the dropout probablity to use for each layer.  The length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        activation_fns: list or object
            the Tensorflow activation function to apply to each layer.  The length of this list should equal
            len(layer_sizes).  Alternatively this may be a single value instead of a list, in which case the
            same value is used for every layer.
        n_classes: int
            the number of classes
        bypass_layer_sizes: list
            the size of each dense layer in the bypass network. The length of this list determines the number of bypass layers.
        bypass_weight_init_stddevs: list or float
            the standard deviation of the distribution to use for weight initialization of bypass layers.
            same requirements as weight_init_stddevs
        bypass_bias_init_consts: list or float
            the value to initialize the biases in bypass layers
            same requirements as bias_init_consts
        bypass_dropouts: list or float
            the dropout probablity to use for bypass layers.
            same requirements as dropouts
        """
        if not isinstance(activation_fns, nn.Module):
            logger.warning(
                "Warning: Activation functions should be of type nn.Module. Using default activation function: ReLU."
            )
            activation_fns = nn.ReLU()

        # The labels are one-hot encoded.
        loss = losses.SoftmaxCrossEntropy()
        output_types = ['prediction', 'loss']
        self.n_classes = n_classes
        self.n_tasks = n_tasks

        model = RobustMultitask(
            n_tasks=self.n_tasks,
            n_features=n_features,
            layer_sizes=layer_sizes,
            mode='classification',
            weight_init_stddevs=weight_init_stddevs,
            bias_init_consts=bias_init_consts,
            weight_decay_penalty=weight_decay_penalty,
            weight_decay_penalty_type=weight_decay_penalty_type,
            activation_fns=activation_fns,
            dropouts=dropouts,
            n_classes=self.n_classes,
            bypass_layer_sizes=bypass_layer_sizes,
            bypass_weight_init_stddevs=bypass_weight_init_stddevs,
            bypass_bias_init_consts=bypass_bias_init_consts,
            bypass_dropouts=bypass_dropouts)
        self.activation_fns = model.activation_fns
        self.dropouts = model.dropouts
        self.shared_layers = model.shared_layers
        self.bypass_layers = model.bypass_layers
        self.output_layers = model.output_layers

        super(RobustMultitaskClassifier,
              self).__init__(model,
                             loss,
                             output_types=output_types,
                             regularization_loss=model.regularization_loss,
                             **kwargs)

    def default_generator(self,
                          dataset,
                          epochs=1,
                          mode='fit',
                          deterministic=True,
                          pad_batches=True):
        """Create a generator that iterates batches for a dataset.

        Subclasses may override this method to customize how model inputs are
        generated from the data.

        Parameters
        ----------
        dataset: Dataset
            the data to iterate
        epochs: int
            the number of times to iterate over the full dataset
        mode: str
            allowed values are 'fit' (called during training), 'predict' (called
            during prediction), and 'uncertainty' (called during uncertainty
            prediction)
        deterministic: bool
            whether to iterate over the dataset in order, or randomly shuffle the
            data for each epoch
        pad_batches: bool
            whether to pad each batch up to this model's preferred batch size

        Returns
        -------
        a generator that iterates batches, each represented as a tuple of lists:
        ([inputs], [outputs], [weights])
        """
        for epoch in range(epochs):
            logger.info("Starting training for epoch %d at %s" %
                        (epoch, datetime.datetime.now().ctime()))
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                if y_b is not None:
                    y_b = to_one_hot(y_b.flatten(), self.n_classes).reshape(
                        -1, self.n_tasks, self.n_classes)
                yield ([X_b], [y_b], [w_b])


class RobustMultitaskRegressor(TorchModel):
    """Implements a neural network for robust multitasking.

    The key idea of this model is to have bypass layers that feed
    directly from features to task output. This might provide some
    flexibility toroute around challenges in multitasking with
    destructive interference.

    References
    ----------
    This technique was introduced in [1]_

    .. [1] Ramsundar, Bharath, et al. "Is multitask deep learning practical for pharma?." Journal of chemical information and modeling 57.8 (2017): 2068-2076.

    """

    def __init__(self,
                 n_tasks,
                 n_features,
                 layer_sizes=[1000],
                 weight_init_stddevs: OneOrMany[float] = 0.02,
                 bias_init_consts: OneOrMany[float] = 1.0,
                 weight_decay_penalty: float = 0.0,
                 weight_decay_penalty_type: str = "l2",
                 dropouts: OneOrMany[float] = 0.5,
                 activation_fns: OneOrMany[ActivationFn] = nn.ReLU(),
                 bypass_layer_sizes=[100],
                 bypass_weight_init_stddevs=[.02],
                 bypass_bias_init_consts=[1.0],
                 bypass_dropouts=[0.5],
                 **kwargs):

        loss = losses.L2Loss()
        output_types = ['prediction']
        n_classes = 1

        model = RobustMultitask(
            n_tasks=n_tasks,
            n_features=n_features,
            layer_sizes=layer_sizes,
            mode='regression',
            weight_init_stddevs=weight_init_stddevs,
            bias_init_consts=bias_init_consts,
            weight_decay_penalty=weight_decay_penalty,
            weight_decay_penalty_type=weight_decay_penalty_type,
            activation_fns=activation_fns,
            dropouts=dropouts,
            n_classes=n_classes,
            bypass_layer_sizes=bypass_layer_sizes,
            bypass_weight_init_stddevs=bypass_weight_init_stddevs,
            bypass_bias_init_consts=bypass_bias_init_consts,
            bypass_dropouts=bypass_dropouts)

        self.shared_layers = model.shared_layers
        self.bypass_layers = model.bypass_layers
        self.output_layers = model.output_layers

        super(RobustMultitaskRegressor,
              self).__init__(model,
                             loss,
                             output_types=output_types,
                             regularization_loss=model.regularization_loss,
                             **kwargs)



from transformers import AutoModel, AutoModelForMaskedLM, AutoTokenizer, AutoConfig
from tqdm import tqdm
from deepchem.models.torch_models import HuggingFaceModel
from typing import Union, Dict, Any


class DeepAbLLM(HuggingFaceModel):
    """Flexible Antibody Language Model for Re-Design of Ab Residues.

    DeepAbLLM is a wrapper class that leverage large language model's (LLMs) learned sequence
    co-dependencies to aid in the (re)-designing anitbody sequences. It supports the instantiation
    of an arbitrary HuggingFace transformer-style model trained on Antibody sequences for the antibody
    sequence redesign, extending an approach introduced in Hie et al's 2023 Nature Biotech paper [1].

    This means the functionality of DeepAbLLM is model architecture-agnostic.

    Currently supports a variety of HuggingFace Protein and Antibody Specific Language Models, including:

    [2] ProtTrans Models (ProtBERT, ProtT5, etc.)
    [3] AbLang
    [4] IgBERT/IgT5
    [5] ESM1b
    [6] ESM1v
    [7] ESM-2

    The model uses single amino acid tokenization to create input tokens for the models from the
    antibody sequences. While most protein models expect spaces in the protein sequences:
        "T H I S I S A P R O T E I N S E Q U E N C E"
    the ESM class of models does not and expects strings of the following format:
        "THISISAPROTEINSEQUENCE"
    Both tokenization schemes are supported by DeepAbLLM by setting the is_esm_variant flag to the
    appropriate value.

    The model supports general pretraining via masked language modeling, domain-adaptive pretraining [8]
    - an additional pretraining step applied to general purpose protein language models for antibody sequences,
    and the finetuning of pre-trained models for regression/classification. To pretrain via masked language
    modeling task, use task = `mlm`, 'regression', or `classification` during initialization.

    Attributes
    ----------
    task: str
        The task the HuggingFaceModel is performing. Default: 'mlm'.
    model_path: str
        The huggingface model path of the pLM.
    n_tasks: int
        Number of tasks for a given model. Default: 1
    is_esm_variant: bool
        Boolean flag to indicate the tokenization scheme.

    Methods
    -------
    __init__(task, model_path, n_tasks)
        Initialize an DeepAbLLM with specified information.
    _mask_seq_pos(position, idx)
        Mask a sequence at a particular index.
    redesign_residue(sequence, residue_index: int, top_k, verbose)
        Mask and unmaks a single residue of a sequence at given index, returning
        top-k plausible amino acid substitutions.
    _optimize_residue_pos(sequence, residue_index, verbose, threshold)
        "Optimizes" a residue position by redesigning it and returning
        only the tokens that rank higher than the original token.
    redesign_sequence(sequence)
        Returns optimized sequences over each residue position with better
        scores than the original sequence.


    Notes
    -----
    (Currently Implements):
    1. Light or Heavy Chain Re-Design at Arbitrary Point
    2. Agnostic to Light or Heavy Chain (Depending on if the specified
                                         model correctly accounts for this)
    (WIP)
    3. Model consumes both epitope and receptor information to influence logits
    (Planned)
    4. Conditional Generation (Auto-Regressive/Iterative Unmasking)

    References
    ----------
    .. [1] Hie, B.L., Shanker, V.R., Xu, D. et al. Efficient evolution of human antibodies
           from general protein language models. Nat Biotechnol 42, 275–283 (2024).
           https://doi.org/10.1038/s41587-023-01763-2

    .. [2] Elnaggar, Ahmed, et al. "Prottrans: Toward understanding the language
           of life through self-supervised learning." IEEE transactions on pattern
           analysis and machine intelligence 44.10 (2021): 7112-7127.

    .. [3] Tobias H Olsen, Iain H Moal, Charlotte M Deane, AbLang: an antibody language
           model for completing antibody sequences, Bioinformatics Advances, Volume 2,
           Issue 1, 2022, vbac046, https://doi.org/10.1093/bioadv/vbac046

    .. [4] Kenlay, H., Dreyer, F. A., Kovaltsuk, A., Miketa, D., Pires, D.,
           & Deane, C. M. (2024). Large scale paired antibody language models.
           arXiv [q-Bio.BM]. Retrieved from http://arxiv.org/abs/2403.17889

    .. [5] Rives A, Meier J, Sercu T, Goyal S, Lin Z, Liu J, Guo D, Ott M, Zitnick CL,
           Ma J, Fergus R. Biological structure and function emerge from scaling
           unsupervised learning to 250 million protein sequences. Proc Natl Acad Sci USA.
           2021 Apr 13;118(15):e2016239118. doi: 10.1073/pnas.2016239118. PMID: 33876751

    .. [6] Meier, J., Rao, R., Verkuil, R., Liu, J., Sercu, T., & Rives, A. (2021).
           Language models enable zero-shot prediction of the effects of mutations
           on protein function. bioRxiv. doi:10.1101/2021.07.09.450648

    .. [7] Zeming Lin et al., Evolutionary-scale prediction of atomic-level
           protein structure with a language model.Science379,1123-1130(2023).
           DOI:10.1126/science.ade2574

    .. [8] Gururangan, S., Marasović, A., Swayamdipta, S., Lo, K., Beltagy, I.,
           Downey, D., & Smith, N. A. (2020). Don’t Stop Pretraining: Adapt
           Language Models to Domains and Tasks. arXiv [Cs.CL].
           Retrieved from http://arxiv.org/abs/2004.10964


    Example Usage:
    --------------
    >>> # Optimize Sequence
    >>> from deepchem.models.torch_models.antibody_modeling import DeepAbLLM
    >>> model_path = 'Rostlab/prot_bert'
    >>> anti_llm = DeepAbLLM(task='mlm', model_path=model_path, n_tasks=1, is_esm_variant=False, device='cpu')
    >>> optimized_sequences = anti_llm.redesign_sequence('GSELTQDPAVSVALGQTVRITCQGDSLRNYYASWYQQKPRQAPVLVFYGKNNRPSGIPDRFSGSSSGNTASLTISGAQAEDEADYYCNSRDSSSNHLVFGGGTKLTVLSQ')
    >>> # Expected Output
    >>> # optimized_sequences[0]
    >>> # (0,'Q','QSETQDPAVSVALGQTVRITCQGDSLRNYYASWYQQKPRQAPVLVFYGKNNRPSGIPDRFSGSSSGNTASLTISGAQAEDEADYYCNSRDSSSNHLVFGGGTKLTVLSQ',0.7314766049385071)
    """

    def __init__(self,
                 task: str = "mlm",
                 model_path: str = 'Rostlab/prot_bert',
                 n_tasks: int = 1,
                 is_esm_variant: bool = False,
                 config: Dict[Any, Any] = {},
                 **kwargs) -> None:
        """
        Parameters
        ----------
        task: str
            The task defines the type of learning task in the model. The supported tasks are
            - `mlm` - masked language modeling commonly used in pretraining
            - `classification` - use it for classification tasks
        model_path: str
            Path to the HuggingFace model; HF Model Hub or local
            - ex: 'Rostlab/prot_bert'
        n_tasks: int
            Number of prediction targets for a multitask learning model
        is_esm_variant: bool
            Flag for proper tokenization (S E Q U E N C E vs SEQUENCE).
        config: dict
            Dictionary of HuggingFace AutoConfig hyper-parameters to update the
            default pretrained model.
        """
        self.n_tasks: int = n_tasks
        tokenizer: AutoTokenizer = AutoTokenizer.from_pretrained(
            model_path, do_lower_case=False)
        model_config: AutoConfig = AutoConfig.from_pretrained(
            pretrained_model_name_or_path=model_path,
            vocab_size=tokenizer.vocab_size)  # type: ignore
        model_config.update(config)  # type: ignore
        self.is_esm_variant: bool = is_esm_variant
        model: Union[AutoModel, AutoModelForMaskedLM]
        if task == "mlm":
            model = AutoModelForMaskedLM.from_config(model_config)
        else:
            model = AutoModel.from_config(model_config)
        super().__init__(
            model=model,  # type: ignore
            task=task,
            tokenizer=tokenizer,  # type: ignore
            **kwargs)

    def _mask_seq_pos(
        self,
        sequence: str,
        idx: int,
    ):
        '''Given an arbitrary antibody sequence with and a seqeunce index,
        convert the residue at that index into the mask token.

        Parameters
        ----------
        sequence: str
            The sequence to be masked at a sepcific residue
        idx: int
            The 0-indexed integer index in which into mask the sequence

        Returns
        -------
        masked_sequence: str
            A nearly identical sequence to the input sequence with the token
            at position idx+1 being the mask token.
        '''
        assert isinstance(idx,
                          int), f"Index must be an int, got type {type(idx)}"
        assert abs(idx) < len(sequence), f"Zero-indexed idx needs to \
            be less than {len(sequence)-1} for sequence of length {len(sequence)}."

        cleaned_sequence = sequence.replace(
            ' ', '')  # Get ride of extraneous spaces if any
        temp_sequence = list(cleaned_sequence)  # Turn the sequence into a list
        temp_sequence[idx] = '*'  # Mask the sequence at idx
        if not self.is_esm_variant:
            masked_sequence = ' '.join(
                temp_sequence
            )  # ProtBERT based models expect 'A M I N O A C I D'
        else:
            masked_sequence = ''.join(
                temp_sequence)  # ESM models expect 'AMINOACID'
        masked_sequence = masked_sequence.replace('*',
                                                  self.tokenizer.mask_token)
        return masked_sequence

    def redesign_residue(self,
                         sequence: str,
                         residue_index: int,
                         top_k: int = 10,
                         verbose: bool = False):
        '''Given a sequence and a residue index, mask and subsequently
        unmask that position, returning the proposed residues and their
        respective scores.

        Parameters
        ----------
        sequence: str
            The antibody sequence to redesign.
        residue_index: int
            The residue index to mask and unmask.
        top_k: int
            The top_k logits to return. Defaults to 10.
        verbose: bool
            If verbose, prints the original sequence and the residue at residue_index
            before designing. Useful for running scripts on clusters.

        Returns
        -------
        sequence_tuples: List[tuple]
            Returns a list of tuples containing the
            (replacement token, full sequence, score) for each unmasked token.

        '''
        masked_sequence = self._mask_seq_pos(sequence, residue_index)
        results = self.fill_mask(masked_sequence,
                                 top_k=top_k)  # List of dictionaries

        sequence_tuples = [(result.get('token_str',
                                       ''), result.get('sequence',
                                                       '').replace(' ', ''),
                            result.get('score', None))
                           for result in results
                           if isinstance(result, dict)]
        if verbose:
            print(
                f"Original Residue at Position {residue_index}: {sequence[residue_index]}\n"
            )
            for i, st in enumerate(sequence_tuples):
                print(f"Redesigned residue {i+1}: {st[0]}, score: {st[-1]}")

        return sequence_tuples

    def _optimize_residue_pos(self,
                              sequence: str,
                              residue_index: int,
                              verbose: bool = False,
                              threshold: float = 0.0,
                              **kwargs):
        '''This is a function to return the optimized residues, as defined
        as the proposed residues that are above a given threshold in probability
        using the masking and unmasking approach. Defualt behaviour returns sequences
        with higher scores than the original sequence.

        Parameters
        ----------
        sequence: str
            Antibody sequence to be optimized at particular index.
        residue_index: int
            Index to optimize input antibody sequence.
        verbose: bool

        Optional:
            top_k: int
                Top K logits to be returned by the redesign_residue method
            threshold: float
                Threshold for probability score

        Returns
        -------
        optimized_sequences: List[tuple]
            Returns list of tuples (token, sequence, score) with higher scores
            than the original and the sequence threshold specified.
        '''
        assert (threshold >= 0) and (
            threshold <=
            1), "Threshold on probability scores should be between 0,1."
        # Sorted List (by score) of redesigned residues
        redesigned_residues = self.redesign_residue(
            sequence, residue_index,
            kwargs['top_k']) if 'top_k' in kwargs else self.redesign_residue(
                sequence, residue_index)

        original_token_str = sequence[residue_index]
        optimized_sequences = []
        for (token_string, full_sequence, score) in redesigned_residues:
            # Check that the token_string is not the same as the original
            if token_string != original_token_str:
                # If it is above a certain probability threshold, append it
                if score > threshold:
                    optimized_sequences += [(token_string, full_sequence, score)
                                           ]
            else:
                # Once the original token is reached break the loop
                break

        return optimized_sequences

    def redesign_sequence(self, sequence: str, **kwargs):
        '''Applies the _optimize_residue_pos function to all sequence positions.

        Parameters
        ----------
        sequence: str
            Antibody sequence to be optimized

        Optional:
            top_k: int
                Top K logits to be returned by the redesign_residue method
            threshold: float
                Threshold for probability score
            verbose: bool
                Flag to print original and redesigned tokens to the stdout.

        Returns
        -------
        redesigned_sequences: List[tuple]
            Returns list of tuples (index, token, sequence, score)
            that have higher scores than the original and are higher than the
            sequence threshold specified.

        '''
        redesigned_sequences = []
        for i in tqdm(range(len(sequence)),
                      desc='Redesigning each residue position'):
            redesigned_sequences += [
                (i,) + x
                for x in self._optimize_residue_pos(sequence, i, **kwargs)
            ]
        return redesigned_sequences
